@@ -56,6 +56,9 @@ extern "C" void unireg_clear(int exit_code)
   DBUG_VOID_RETURN;
 }
 
+
+static my_bool mysql_embedded_init= 0;
+
 /*
   Wrapper error handler for embedded server to call client/server error 
   handler based on whether thread is in client/server context
@@ -520,6 +523,7 @@ int init_embedded_server(int argc, char **argv, char **groups)
   const char *fake_groups[] = { "server", "embedded", 0 };
   my_bool acl_error;
 
+  DBUG_ASSERT(mysql_embedded_init == 0);
   embedded_print_errors= 1;
   if (my_thread_init())
     return 1;
@@ -623,7 +627,8 @@ int init_embedded_server(int argc, char **argv, char **groups)
 
   (void) thr_setconcurrency(concurrency);	// 10 by default
 
-  start_handle_manager();
+  if (flush_time && flush_time != ~(ulong) 0L)
+    start_handle_manager();
 
   // FIXME initialize binlog_filter and rpl_filter if not already done
   //       corresponding delete is in clean_up()
@@ -640,15 +645,20 @@ int init_embedded_server(int argc, char **argv, char **groups)
   }
 
   execute_ddl_log_recovery();
+  mysql_embedded_init= 1;
   return 0;
 }
 
 void end_embedded_server()
 {
-  my_free(copy_arguments_ptr);
-  copy_arguments_ptr=0;
-  clean_up(0);
-  clean_up_mutexes();
+  if (mysql_embedded_init)
+  {
+    my_free(copy_arguments_ptr);
+    copy_arguments_ptr=0;
+    clean_up(0);
+    clean_up_mutexes();
+    mysql_embedded_init= 0;
+  }
 }
 
 
@@ -658,7 +668,7 @@ void init_embedded_mysql(MYSQL *mysql, int client_flag)
   thd->mysql= mysql;
   mysql->server_version= server_version;
   mysql->client_flag= client_flag;
-  init_alloc_root(&mysql->field_alloc, 8192, 0, MYF(0));
+  init_alloc_root(&mysql->field_alloc, "fields", 8192, 0, MYF(0));
 }
 
 /**
@@ -695,8 +705,7 @@ void *create_embedded_thd(int client_flag)
   thd->client_capabilities= client_flag;
   thd->real_id= pthread_self();
 
-  thd->db= NULL;
-  thd->db_length= 0;
+  thd->db= null_clex_str;
 #ifndef NO_EMBEDDED_ACCESS_CHECKS
   thd->security_ctx->db_access= DB_ACLS;
   thd->security_ctx->master_access= ~NO_ACCESS;
@@ -734,7 +743,7 @@ emb_transfer_connect_attrs(MYSQL *mysql)
     ptr= buf= (uchar *) my_alloca(length + 9);
     send_client_connect_attrs(mysql, buf);
     net_field_length_ll(&ptr);
-    PSI_THREAD_CALL(set_thread_connect_attrs)((char *) ptr, length, thd->charset());
+    PSI_CALL_set_thread_connect_attrs((char *) ptr, length, thd->charset());
     my_afree(buf);
   }
 #endif
@@ -744,7 +753,7 @@ emb_transfer_connect_attrs(MYSQL *mysql)
 int check_embedded_connection(MYSQL *mysql, const char *db)
 {
   int result;
-  LEX_STRING db_str = { (char*)db, db ? strlen(db) : 0 };
+  LEX_CSTRING db_str = { db, safe_strlen(db) };
   THD *thd= (THD*)mysql->thd;
 
   /* the server does the same as the client */
@@ -975,7 +984,7 @@ int Protocol::begin_dataset()
     return 1;
   alloc= &data->alloc;
   /* Assume rowlength < 8192 */
-  init_alloc_root(alloc, 8192, 0, MYF(0));
+  init_alloc_root(alloc, "protocol", 8192, 0, MYF(0));
   alloc->min_malloc= sizeof(MYSQL_ROWS);
   return 0;
 }
@@ -1038,7 +1047,7 @@ bool Protocol::send_result_set_metadata(List<Item> *list, uint flags)
   while ((item= it++))
   {
     Send_field server_field;
-    item->make_field(thd, &server_field);
+    item->make_send_field(thd, &server_field);
 
     /* Keep things compatible for old clients */
     if (server_field.type == MYSQL_TYPE_VARCHAR)
@@ -1048,12 +1057,14 @@ bool Protocol::send_result_set_metadata(List<Item> *list, uint flags)
                                   strlen(server_field.db_name), cs, thd_cs);
     client_field->table= dup_str_aux(field_alloc, server_field.table_name,
                                      strlen(server_field.table_name), cs, thd_cs);
-    client_field->name= dup_str_aux(field_alloc, server_field.col_name,
-                                    strlen(server_field.col_name), cs, thd_cs);
+    client_field->name= dup_str_aux(field_alloc, server_field.col_name.str,
+                                    server_field.col_name.length, cs, thd_cs);
     client_field->org_table= dup_str_aux(field_alloc, server_field.org_table_name,
                                          strlen(server_field.org_table_name), cs, thd_cs);
-    client_field->org_name= dup_str_aux(field_alloc, server_field.org_col_name,
-                                        strlen(server_field.org_col_name), cs, thd_cs);
+    client_field->org_name= dup_str_aux(field_alloc,
+                                        server_field.org_col_name.str,
+                                        server_field.org_col_name.length,
+                                        cs, thd_cs);
     if (item->charset_for_protocol() == &my_charset_bin || thd_cs == NULL)
     {
       /* No conversion */
