@@ -566,7 +566,7 @@ int _my_b_read(IO_CACHE *info, uchar *Buffer, size_t Count)
   }
   res= info->read_function(info, Buffer, Count);
   if (res && info->error >= 0)
-    info->error+= left_length; /* update number or read bytes */
+    info->error+= (int)left_length; /* update number or read bytes */
   return res;
 }
 
@@ -1603,7 +1603,7 @@ int _my_b_async_read(IO_CACHE *info, uchar *Buffer, size_t Count)
     Buffer+=length;
     Count-=length;
     left_length+=length;
-    info->read_end=info->read_pos+read_length;
+    info->read_end=info->rc_pos+read_length;
     info->read_pos+=length;
   }
   else
@@ -1935,13 +1935,12 @@ int my_b_flush_io_cache(IO_CACHE *info, int need_append_buffer_lock)
     {
       if (append_cache)
       {
-
         if (mysql_file_write(info->file, info->write_buffer, length,
                              info->myflags | MY_NABP))
+        {
           info->error= -1;
-        else
-          info->error= 0;
-
+          DBUG_RETURN(-1);
+        }
         info->end_of_file+= info->write_pos - info->append_read_pos;
         info->append_read_pos= info->write_buffer;
         DBUG_ASSERT(info->end_of_file == mysql_file_tell(info->file, MYF(0)));
