@@ -26,13 +26,13 @@
   Used instead of FILE when reading or writing whole files.
   This will make mf_rec_cache obsolete.
   One can change info->pos_in_file to a higher value to skip bytes in file if
-  also info->read_pos is set to info->read_end.
+  also info->rc_pos is set to info->rc_end.
   If called through open_cached_file(), then the temporary file will
   only be created if a write exeeds the file buffer or if one calls
   flush_io_cache().  
 */
 
-#include <my_global.h>
+#include "mariadb.h"
 #include "sql_priv.h"
 #include "sql_class.h"                          // THD
 #ifdef HAVE_REPLICATION
@@ -58,12 +58,12 @@ int _my_b_net_read(IO_CACHE *info, uchar *Buffer, size_t)
   if (!info->end_of_file)
     DBUG_RETURN(1);	/* because my_b_get (no _) takes 1 byte at a time */
   read_length= my_net_read_packet(net, 0);
-  if (read_length == packet_error)
+  if (unlikely(read_length == packet_error))
   {
     info->error= -1;
     DBUG_RETURN(1);
   }
-  if (read_length == 0)
+  if (unlikely(read_length == 0))
   {
     info->end_of_file= 0;			/* End of file from client */
     DBUG_RETURN(1);
