@@ -34,7 +34,7 @@ Created 03/11/2014 Shaohua Wang
 #include <vector>
 
 /** Innodb B-tree index fill factor for bulk load. */
-extern	long	innobase_fill_factor;
+extern	uint	innobase_fill_factor;
 /** whether to reduce redo logging during ALTER TABLE */
 extern	my_bool	innodb_log_optimize_ddl;
 
@@ -87,7 +87,7 @@ public:
 		m_err(DB_SUCCESS)
 	{
 		ut_ad(!dict_index_is_spatial(m_index));
-		ut_ad(!dict_table_is_temporary(m_index->table));
+		ut_ad(!m_index->table->is_temporary());
 	}
 
 	/** Deconstructor */
@@ -292,7 +292,8 @@ public:
 		ut_ad(!dict_index_is_spatial(index));
 #ifdef UNIV_DEBUG
 		if (m_flush_observer)
-		fil_space_inc_redo_skipped_count(m_index->space);
+		my_atomic_addlint(&m_index->table->space->redo_skipped_count,
+				  1);
 #endif /* UNIV_DEBUG */
 	}
 
@@ -301,7 +302,8 @@ public:
 	{
 #ifdef UNIV_DEBUG
 		if (m_flush_observer)
-		fil_space_dec_redo_skipped_count(m_index->space);
+		my_atomic_addlint(&m_index->table->space->redo_skipped_count,
+				  ulint(-1));
 #endif /* UNIV_DEBUG */
 	}
 
